#********************************
# Implementor: Raghvendra Rohit
# Contact: rsrohit@uwaterloo.ca 
#********************************

from gurobipy import *

class wage :
	def __init__(self, rounds, ft, sut, mid, nw):
		self.rounds = rounds 											# Number of rounds
		self.ft = ft													# Feedback taps including last sbox
		self.sut = sut 													# Sbox and updated taps (without last and middle wgp sbox)
		self.nw = nw													# Number of words in LFSR
		self.mid = mid 													# Tap position of middle wgp sbox
		self.file_model = "wage_" + str(self.rounds)  + ".lp"
		self.file_binary = "wage_binary" + str(self.rounds)  +".txt"
		fp = open(self.file_model,"w")
		fp.close()
		fp = open(self.file_binary,"w")
		fp.close()

	def create_variables(self, x, r, len):
		variable = []
		for i in range(len):
			variable.append(x + "_" + str(r) + "_" + str(i))
		return variable

	def create_dvariables(self, x, r):
		variable = []
		variable.append(x + "_" + str(r))
		return variable

	def xor_two_words(self, x_in0, x_in1, x_out, r, ind):
		
		D = self.create_dvariables('d'+str(ind), r)
		
		fp = open(self.file_binary, "a")
		fp.write(D[0] + str("\n"))
		fp.close()
		
		fp = open(self.file_model, "a")
	
		fp.write(x_in0 + " + " + x_in1 + " + " + x_out + " - 2 " + D[0] + " >= 0 \n")
		fp.write(D[0] + " - " + x_in0 + " >= 0\n")
		fp.write(D[0] + " - " + x_in1 + " >= 0\n")
		fp.write(D[0] + " - " + x_out + " >= 0\n")

		fp.close()

	def fb_xor(self, X, Y, r):											
		
		D = self.create_dvariables('d'+str(self.nw-1), r)

		fp = open(self.file_binary, "a")
		fp.write(D[0] + str("\n"))
		fp.close()

		fp = open(self.file_model, "a")
		temp = []

		for t in X: 
			temp.append(t)

		temp.append(Y[0])
		temp = " + ".join(temp)
		temp = temp + " - 2 " + D[0] + " >= 0"
		fp.write(temp + "\n")

		for t in X:
			fp.write(D[0] + " - " + t + " >= 0\n")
		fp.write(D[0] + " - " + Y[0] + " >= 0\n")
		fp.close()

	def shift_state(self, X, Y):

		for i in range(self.nw - 1):
			X[i] = X[i+1]
		X[self.nw-1] = Y
		return X
		
	def round_function(self):

		X = self.create_variables('x', 0, self.nw)				
	
		fp = open(self.file_binary, "a")
		for i in range(self.nw):
			fp.write(X[i] + "\n")
		fp.close()

		for i in range(self.rounds):

			Y = self.create_variables('x', i+1, self.nw)
			
			fp = open(self.file_binary,'a')
			fp.write(Y[-1] + "\n")
			for t in self.sut:
				fp.write(Y[t[1]] + "\n")
			fp.write(Y[self.mid+1] + "\n")

			fp.close()

			self.fb_xor([X[t] for t in self.ft], [Y[-1]], i)
			
			for t in self.sut:
				self.xor_two_words(X[t[0]], X[t[1]], Y[t[1]], i, t[1])

			self.xor_two_words(X[self.mid], X[self.mid + 1], Y[self.mid+1], i, self.mid+1)

			
			for t in self.sut:
				X[t[1]] = Y[t[1]]

			X[self.mid + 1] = Y[self.mid  + 1]
			X = self.shift_state(X, Y[-1])
		
		# Case II : Output difference at rate positions only
		'''
		rate_index = [8,9,15,16,18,27,28,34,35,36]
		temp = []
		Z = list(set(i for i in range(self.nw))-set(rate_index))
		fp = open(self.file_model, "a")
		for i in rate_index:
			temp.append(X[i])
		temp = " + ".join(temp)
		temp = temp + " >= 1 \n"
		fp.write(temp) 

		for z in Z:
			fp.write(X[z] + " = 0 \n")
		fp.close()
		'''

			
	def init(self):

		fp = open(self.file_model, "a")
		fp.write("Minimize\n")
		
		temp = []
		X = self.create_variables('x', 0, self.nw)

		for t in self.sut:
			temp.append(X[t[0]])
		temp.append(X[self.mid])
		temp.append(X[self.nw-1])

		for i in range(self.rounds-1):
			
			Y = self.create_variables('x', i+1, self.nw)
			
			for t in self.sut:
				X[t[1]] = Y[t[1]]
			X[self.mid + 1] = Y[self.mid  + 1]
			X = self.shift_state(X, Y[-1])
			for t in self.sut:
				temp.append(X[t[0]])
			temp.append(X[self.mid])
			temp.append(X[self.nw-1])
			
		temp = " + ".join(temp)
		fp.write(temp + "\n")
		fp.write("Subject To \n")

		temp = []

		## Case I : No constraints on the positions of input difference
		
		for i in range(self.nw):
			temp.append('x_0_'+str(i))
		temp = " + ".join(temp)
		temp = temp + " >= 1 \n"
		fp.write(temp) 
		

		## Case II: Input difference at rate positions only
		'''
		rate_index = [8,9,15,16,18,27,28,34,35,36]
		Z = list(set(i for i in range(self.nw))-set(rate_index))
		for i in rate_index:
			temp.append('x_0_'+str(i))
		temp = " + ".join(temp)
		temp = temp + " >= 1 \n"
		fp.write(temp) 

		for z in Z:
			fp.write('x_0_'+str(z) + " = 0 \n")
		'''
		
		fp.close()
	
	def make_model(self):
		
		self.init()
		self.round_function()
		fileobj = open(self.file_model, "a")
		fileobj1 = open(self.file_binary, "r")
		fileobj.write("Binary\n")
		for line in fileobj1:
			fileobj.write(line)
		fileobj.write("END\n")
		fileobj.close()
		fileobj1.close()

	def solve_model(self):
		
		m = read(self.file_model)
		m.optimize() 
		if(m.Status ==2):
			obj = m.getObjective()
			return (obj.getValue())
		